
banner() {
	local banner=${cefsroot}/lib/self-heal.banner
	test -f $banner && cat $banner
}

debug() {
	if [ -n "$DEBUG" ] ; then
		echo -e "DEBUG:" "$@" 1>&2
	fi
}

find_cefs_root () {
	df $1 | tail -n1 | awk '{ print $1 }' | sed s/cefs://
}

inner_root() {
	df $1 | tail -n1 |  awk '{ print $NF }'
}

injuries() {
	find "$@"/.__.cefs./.root/.quarantine -name \*.cqi | grep -v -e 'db-journal' -e '.db'
}

injured_file() {
	inner_root="$1"
	shift
	dirname "$@" | sed s,$inner_root/.__.cefs./.root/.quarantine,,
}

injury_type() {
	basename "$@" .cqi
}


search_for_unaffected() {
	local patient="$@" second_ext orig_guess base_f
	case $patient in
	*.*.*)
		second_ext=.$(echo "${patient}" | awk -F. '{ print $NF }')
		orig_guess="$(dirname $patient)"/$(basename "${patient}" $second_ext)
		base_f=$(basename "${patient}")
		;;
	*|*.*)
		second_ext=""
		orig_guess="${patient}"
		base_f=$(basename "${patient}")
		;;
	esac
	SECOND_EXT=${SECOND_EXT:+${SECOND_EXT}:}${second_ext}
	local inj="${ROOT}/${patient}"
	debug "ORIG_GUESS: ${orig_guess}"
	ls -1tr ${mem_root}/[1-9]*/"${orig_guess}" 2>/dev/null | (while read c ; do
		if test -f "${inj}"; then
			if ! diff "${c}" "${inj}" 2>&1 >/dev/null; then
				echo "${c}|${orig_guess}|${inj}"
				return 0
			fi
		else
			if ! diff "${c}" "${ROOT}${orig_guess}" 2>&1 >/dev/null ; then
				echo "${c}|${orig_guess}|${ROOT}${orig_guess}"
				return 0
			fi
		fi
		debug "Found association for ${orig_guess} at ${c}"
	done; return 1)
}

clean_bill() {
	local cqi="$@"
	debug "Cleanup: ${cq}"
	dry_run mv -f "${cqi}" "$(dirname $cqi)/$(basename ${cqi} .cqi)._qi"
}

heal() {
	local f="$@"
	local ua=$(search_for_unaffected "$f")
	local source=$(echo $ua | cut -d'|' -f1)
	local dest=$(echo $ua | cut -d'|' -f2)
	local inj=$(echo $ua | cut -d'|' -f3)
	debug "source: \"$source\""
	debug "dest: \"$dest\""
	debug "inj: \"$inj\""
	if [ -f "${source}" ] ; then
		debug cp -fa "${source}" "${ROOT}/${dest}"
		dry_run cp -fa "${source}" "${ROOT}/${dest}"
		if [ -f "${inj}" ] ; then
			debug rm -f "${inj}"
			dry_run rm -f "${inj}"
		fi
	else
		MISSING=${MISSING:+${MISSING}:}${f}
		return 1
	fi
	return 0
}

build_queue() {

	debug "Building Queue..."

	local count=$(injuries $inner_root | wc -l)
	echo "Found $count injuries"

	for i in $(injuries $inner_root); do
		local f=$(injured_file "$inner_root" "$i")
		local t=$(injury_type "$i")
		lecho "${ROOT}/$f..."
		if heal "$f" && clean_bill "$i"  ; then
			lecho "healed"
		else
			lecho "queued for deep heal"
		fi
	done
	echo
}

mem_down() {
	umount $mem_root/* 2>/dev/null
	rmdir $mem_root/* 2>/dev/null
}

mem_up() {
	local start=$1 finish=$2
	local range=$(eval echo `echo {${start}..${finish}}`)
	if test $(basename "$mem_root") != "mem" ; then
		debug "There is a a bug"
		exit -1
	fi
	for m in $range; do
		cefs rectify $path memory $m 2>&1 >/dev/null && lecho "Priming memory ${m} of ${finish}..."
	done
	echo "Ready!"
}

deep_heal() {
	local inj="$1"
	ls -1t ${mem_root}/[1-9]*/"${inj}" 2>/dev/null | (while read c ; do
		if ! test -f "${c}"; then
			continue
		fi
		ft=$(file -i -b -N --mime-type $c | sed s/\;.*//)
		if [ "$ft" != "application/octet-stream" ] ; then
			lecho Healing $inj
			dry_run cp -a "$c"  "${inner_root}/${inj}"
			return 0
		fi
	done)
	return 1
}


pause_fs_procs() {
	local TARGET_DIR=${1:-${ROOT:-/home/WIN_HOMES}}
	info "Stopping services using ${TARGET_DIR}"
	systemctl stop smbd winbind nmbd
	if [ -f ${TARGET_DIR}/.cellfs ]; then
		sleep 1
		pids_lsof=$(lsof -t ${TARGET_DIR}/)
		if [ -n "${pid_lsof}" ]; then
			echo shutting down ${pids_lsof}
			kill ${pid_lsof}
			echo "waiting for 2 seconds"
			sleep 2
		fi
		pids_lsof=$(lsof -t ${TARGET_DIR}/)
		if [ -n "${pid_lsof}" ]; then
			echo "Killing ${pid_lsof}"
			kill -9 ${pid_lsof}
			sleep 1
		fi
	fi
}


resume_fs_procs() {
	info "Restarting services using ${ROOT}"
	systemctl start smbd nmbd winbind
}


settle_down() {
	local root=${1:-${ROOT:-/home/WIN_HOMES}}
	old_mem=0
	count=0
	loop=0
	while : ; do
		loop=$((loop + 1))
		mem=$(cefs rectify ${root} -la | tail -n1 | awk '{ mem=NF-2;  print $mem }')
		lecho "Waiting for ${root} to settle down...$((20 - loop)): "
		if [ "$mem" = "$old_mem" ] ; then
			count=$((count + 1))
			if [ $count -gt 5 ] ; then
				echo "Settled!"
				return 0
			fi
		fi
		old_mem=${mem}
		if [ $loop -gt 20 ] ; then
			echo 1>&2
			die 0 "File-system not settled yet, wait for a bit and try again"
		fi
	done
	exit 0
}

main() {
	init_vars
	ROOT=${1:-/home/WIN_HOMES}
	if [ ! -f ${ROOT}/.cellfs ] ; then
		die 1 "Invalid partition ${ROOT}"
	fi


	path="$ROOT"
	pause_fs_procs

	cefs_root=$(find_cefs_root $path)
	debug $cefs_root
	inner_root=$(inner_root $cefs_root)
	debug $inner_root
	mem_root=$(dirname $inner_root)/mem
	debug $mem_root

	settle_down


	local max=800
	local finish=$(cefs rectify $path -la 2>&1 | tail -n1 | awk '{ print $4 }')
	local start=$((finish - max))
	if test "$start" -lt "1" ; then
		start=1
	fi
	# range=$(eval echo `echo {${start}..${finish}}`)
	mem_up $start $finish $ROOT
	build_queue
	SECOND_EXT=${SECOND_EXT:-.cryptojack}
	debug "Unhealed files:" ${MISSING} 1>&2
	debug "Suspicious files: ${SECOND_EXT}" 1>&2
	info "Deep healing missing files"
	echo ${MISSING} | sed s/:/\\n/g | while read m ; do
		deep_heal $m
	done
	find ${inner_root} -name \*${SECOND_EXT} -printf "%p\n"  | while read p; do
		lecho "[INFO] Removing dead zombie: $(basename ${p})"
		if test -f "${p}" ; then
			rm -f "${p}"
		elif test -d "${p}" ; then
			rm -rf "${p}"
		else
			debug "Weird ${p}"
		fi
	done
	echo
	mem_down
	resume_fs_procs
}


main "$@"